#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <memory.h>

#ifndef FALSE
# undef FALSE
# define FALSE 0
#endif

#ifndef TRUE
# undef TRUE
# define TRUE (!(FALSE))
#endif

#define U32SWAP(N) ((N >> 24) | ((N << 8) & 0xFF0000) | ((N >> 8) & 0xFF00) | ((N << 24) & 0xFF000000))
#define CARY_MAGIC 0x59524143
#define CCB_MAGIC 0x20424343
#define PDAT_MAGIC 0x54414450
#define PDAT_SECTION_MAX_SIZE 0x100000
#define PLUT_MAGIC 0x54554C50
#define PLUT_SECTION_MIN_SIZE 0x40
#define PLUT_SECTION_MAX_SIZE UINT16_MAX
#define PLUT_PLACEHOLDER_VALUE 0x20000000
#define CCB_SECTION_SIZE 0x50000000
#define CCB_OFFSET_CORRECT_DIFF (sizeof(struct ccb_header_raw_part))
#define CEL_FILE_NAME_ARY_PREFIX "ary_"
#define CEL_FILE_NAME_ARY_DELIM "."
#define CEL_FILE_NAME_PREFIX "img_"
#define CEL_FILE_NAME_NUM_MAX_LEN 10
#define CEL_FILA_NAME_SUFFIX ".cel"
#define CEL_FILE_NAME_SIZE (sizeof(CEL_FILE_NAME_PREFIX) - 1 + sizeof(CEL_FILE_NAME_ARY_DELIM) - 1 + (CEL_FILE_NAME_NUM_MAX_LEN * 2) + sizeof(CEL_FILE_NAME_ARY_PREFIX) - 1 + sizeof(CEL_FILA_NAME_SUFFIX))

typedef uint8_t BOOL;

struct __attribute__((packed)) cary_header {
  uint32_t magic;
  uint32_t ccb_section_start_offset;
  uint32_t file_size;
  uint32_t zero_placeholder;
};

struct __attribute__((packed)) point {
  uint32_t x;
  uint32_t y;
};

struct ccb_offsets_col {
  uint32_t self;
  uint32_t pdat;
  uint32_t plut;
  uint32_t pdat_size;
  uint32_t plut_size;
};

struct ccb_offsets_list {
  uint32_t count;
  struct ccb_offsets_col *offsets_col;
};

struct __attribute__((packed)) ccb_header_raw_part {
  uint32_t flags;
  uint32_t offset_next_ccb;
  uint32_t offset_pdat;
  uint32_t offset_plut;
  struct point pos;
  struct point hd;
  struct point vd;
  struct point hdd;
  uint32_t pixc;
  uint32_t pre0;
  uint32_t pre1;
  struct point size;
};

struct ccb_header_raw_part_list {
  struct ccb_header_raw_part *ccb_header_raw_part;
  struct ccb_header_raw_part_list *next;
};

struct __attribute__((packed)) ccb_header {
  uint32_t magic;
  uint32_t section_size;
  uint32_t zero_placeholder;
  struct ccb_header_raw_part raw_part;
};

struct __attribute__((packed)) ccb_header_pdat {
  uint32_t magic;
  uint32_t section_size;
};

struct __attribute__((packed)) ccb_header_plut {
  uint32_t magic;
  uint32_t section_size;
  uint32_t placeholder;
};

static uint32_t current_ary_num = 0;
static uint32_t current_cel_num = 0;

static void cary_get_max_ccb_count_recurse(uint32_t *ccb_count, const uint8_t *const cary_data, const uint32_t offset) {
  const struct ccb_header_raw_part *const ccb_header_raw_part = (const struct ccb_header_raw_part *const)(cary_data + offset);

  *ccb_count += 1;

  if (ccb_header_raw_part->offset_next_ccb > 0) {
    cary_get_max_ccb_count_recurse(ccb_count, cary_data, U32SWAP(ccb_header_raw_part->offset_next_ccb));
  }
}

static const uint32_t cary_get_max_ccb_count(const struct cary_header *const cary_header) {
  uint32_t ccb_offset_index;
  uint32_t ccb_count = 0;
  const uint32_t *const ccb_offsets = (const uint32_t *const)(cary_header + 1);

  for (ccb_offset_index = 0;; ccb_offset_index++) {
    const uint32_t ccb_offset = ccb_offsets[ccb_offset_index];

    if (ccb_offset > 0) {
      cary_get_max_ccb_count_recurse(&ccb_count, (const uint8_t *const)cary_header, U32SWAP(ccb_offset));
    } else {
      break;
    }
  }

  return ccb_count;
}

struct ccb_offsets_list *ccb_offsets_list_create(const struct cary_header *const cary_header) {
  const uint32_t count = cary_get_max_ccb_count(cary_header);
  struct ccb_offsets_list *list = malloc(sizeof(*list));
  const size_t offsets_col_size = sizeof(*list->offsets_col) * count;

  list->count = count;
  list->offsets_col = malloc(offsets_col_size);

  memset((void *)list->offsets_col, 0, offsets_col_size);

  return list;
}

static BOOL ccb_offsets_list_exists(struct ccb_offsets_list *list, struct ccb_offsets_col *offset_col) {
  uint32_t i;

  for (i = 0; i < list->count; i++) {
    const struct ccb_offsets_col *const ccb_offsets_col = list->offsets_col + i;

    if (ccb_offsets_col->self == offset_col->self || ccb_offsets_col->pdat == offset_col->pdat) {
      return TRUE;
    }
  }

  return FALSE;
}

static void ccb_offsets_list_fill_recurse(struct ccb_offsets_list *list, const uint8_t *const cary_data, uint32_t *count, const uint32_t offset) {
  struct ccb_offsets_col new_offset_col;
  const struct ccb_header_raw_part *const ccb_header_raw_part = (const struct ccb_header_raw_part *const)(cary_data + offset);
  const uint32_t ccb_header_next_offset = U32SWAP(ccb_header_raw_part->offset_next_ccb);

  new_offset_col.self = offset;
  new_offset_col.pdat = U32SWAP(ccb_header_raw_part->offset_pdat);
  new_offset_col.plut = U32SWAP(ccb_header_raw_part->offset_plut);

  if (ccb_offsets_list_exists(list, &new_offset_col) == FALSE) {
    struct ccb_offsets_col *offset_col = list->offsets_col + (*count);

    memcpy((void *)offset_col, (const void *)&new_offset_col, sizeof(*offset_col));

    *count += 1;
  }

  if (ccb_header_next_offset > 0) {
    ccb_offsets_list_fill_recurse(list, cary_data, count, ccb_header_next_offset);
  }
}

static void ccb_offsets_list_fill(struct ccb_offsets_list *list, const struct cary_header *const cary_header) {
  uint32_t i;
  uint32_t count = 0;
  const uint8_t *const cary_data = (const uint8_t *const)cary_header;
  const uint32_t *const offsets = (const uint32_t *const)(cary_header + 1);

  for (i = 0;; i++) {
    const uint32_t offset = U32SWAP(offsets[i]);

    if (offset > 0) {
      ccb_offsets_list_fill_recurse(list, cary_data, &count, offset);
    } else {
      break;
    }
  }

  list->count = count;
}

static const uint32_t ccb_offsets_list_get_min_plut(const struct ccb_offsets_list *const list) {
  uint32_t i;
  uint32_t min_plut_offset = UINT32_MAX;

  for (i = 0; i < list->count; i++) {
    const struct ccb_offsets_col *const offsets_col = list->offsets_col + i;

    if (offsets_col->plut < min_plut_offset) {
      min_plut_offset = offsets_col->plut;
    }
  }

  return min_plut_offset;
}

static void ccb_offsets_list_fill_sizes(struct ccb_offsets_list *list, const uint32_t total_cary_size) {
  uint32_t i;
  struct ccb_offsets_col *offsets_col;
  const uint32_t last_index = list->count - 1;
  const uint32_t min_plut_offset = ccb_offsets_list_get_min_plut(list);

  for (i = 0; i < last_index; i++) {
    const struct ccb_offsets_col *const offsets_col_next = list->offsets_col + i + 1;

    offsets_col = list->offsets_col + i;

    if (offsets_col->pdat < offsets_col_next->pdat) {
      offsets_col->pdat_size = offsets_col_next->pdat - offsets_col->pdat;
    } else {
      offsets_col->pdat_size = 0;
    }

    if (offsets_col->plut < offsets_col_next->plut) {
      offsets_col->plut_size = offsets_col_next->plut - offsets_col->plut;
    } else {
      offsets_col->plut_size = 0;
    }
  }

  offsets_col = list->offsets_col + last_index;
  offsets_col->pdat_size = min_plut_offset - offsets_col->pdat;
  offsets_col->plut_size = total_cary_size - offsets_col->plut;
}

static int ccb_offsets_list_sort_compare(const void *lhs, const void *rhs) {
  const struct ccb_offsets_col *const lhs_offsets_col = lhs;
  const struct ccb_offsets_col *const rhs_offsets_col = rhs;

  return lhs_offsets_col->pdat - rhs_offsets_col->pdat;
}

static void ccb_offsets_list_sort(struct ccb_offsets_list *list) {
  qsort((void *)list->offsets_col, list->count, sizeof(*list->offsets_col), ccb_offsets_list_sort_compare);
}

static void ccb_offsets_col_print(const struct ccb_offsets_col *const offsets_col) {
  printf("CCB "
    "offset_self=%u, "
    "offset_pdat=%u, "
    "offset_plut=%u, "
    "size_pdat=%u, "
    "size_plut=%u\n",
      offsets_col->self,
      offsets_col->pdat,
      offsets_col->plut,
      offsets_col->pdat_size,
      offsets_col->plut_size
  );
}

/*
static void ccb_offsets_list_print(const struct ccb_offsets_list *const list) {
  uint32_t i;

  for (i = 0; i < list->count; i++) {
    ccb_offsets_col_print(list->offsets_col + i);
  }
}
*/

static void ccb_offsets_list_destroy(struct ccb_offsets_list *list) {
  free((void *)list->offsets_col);
  free((void *)list);
}

static void process_ccb(const uint8_t *const cary_data, const struct ccb_offsets_col *const ccb_offsets_col) {
  int ccb_file_fd;
  uint8_t *ccb_file_contents;
  char cel_file_name[CEL_FILE_NAME_SIZE];

  struct ccb_header *file_ccb_header;
  struct ccb_header_pdat *file_ccb_header_pdat;
  uint8_t *file_pdat_data;
  struct ccb_header_plut *file_ccb_header_plut;
  uint8_t *file_plut_data;

  const uint32_t pdat_section_size = ccb_offsets_col->pdat_size + sizeof(*file_ccb_header_pdat);
  const uint32_t plut_section_size = ccb_offsets_col->plut_size + sizeof(*file_ccb_header_plut);
  size_t ccb_file_size = sizeof(*file_ccb_header) + pdat_section_size + plut_section_size;

  if (ccb_offsets_col->pdat_size >= PDAT_SECTION_MAX_SIZE) {
    printf("CCB invalid PDAT section size\n");
    return;
  } else if (ccb_offsets_col->plut_size >= PLUT_SECTION_MAX_SIZE) {
    printf("CCB invalid PLUT section size %u\n", ccb_offsets_col->plut_size);
    return;
  } else if (ccb_offsets_col->plut_size < PLUT_SECTION_MIN_SIZE) {
    printf("CCB no PLUT section\n");
    ccb_file_size -= plut_section_size;
  } else {
    ccb_offsets_col_print(ccb_offsets_col);
  }

  snprintf(cel_file_name, CEL_FILE_NAME_SIZE, CEL_FILE_NAME_ARY_PREFIX "%010u" CEL_FILE_NAME_ARY_DELIM CEL_FILE_NAME_PREFIX "%010u" CEL_FILA_NAME_SUFFIX, current_ary_num, ++current_cel_num);

  ccb_file_fd = open(cel_file_name, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
  ftruncate(ccb_file_fd, ccb_file_size);
  lseek(ccb_file_fd, 0, SEEK_SET);

  ccb_file_contents = mmap(NULL, ccb_file_size, PROT_READ | PROT_WRITE, MAP_SHARED, ccb_file_fd, 0);

  file_ccb_header = (struct ccb_header *)ccb_file_contents;
  file_ccb_header_pdat = (struct ccb_header_pdat *)(file_ccb_header + 1);
  file_pdat_data = (uint8_t *)(file_ccb_header_pdat + 1);
  file_ccb_header_plut = (struct ccb_header_plut *)(file_pdat_data + ccb_offsets_col->pdat_size);
  file_plut_data = (uint8_t *)(file_ccb_header_plut + 1);

  memcpy((void *)&file_ccb_header->raw_part, (const void *)(cary_data + ccb_offsets_col->self), sizeof(file_ccb_header->raw_part));
  file_ccb_header->magic = CCB_MAGIC;
  file_ccb_header->section_size = U32SWAP(sizeof(*file_ccb_header));
  file_ccb_header->zero_placeholder = 0;
  file_ccb_header->raw_part.offset_next_ccb = 0;
  file_ccb_header->raw_part.offset_pdat = 0;
  file_ccb_header->raw_part.offset_plut = 0;

  file_ccb_header_pdat->magic = PDAT_MAGIC;
  file_ccb_header_pdat->section_size = U32SWAP(pdat_section_size);
  memcpy((void *)file_pdat_data, (const void *)(cary_data + ccb_offsets_col->pdat), ccb_offsets_col->pdat_size);

  if (ccb_offsets_col->plut_size >= PLUT_SECTION_MIN_SIZE) {
    file_ccb_header_plut->magic = PLUT_MAGIC;
    file_ccb_header_plut->section_size = U32SWAP(plut_section_size);
    file_ccb_header_plut->placeholder = PLUT_PLACEHOLDER_VALUE;
    memcpy((void *)file_plut_data, (const void *)(cary_data + ccb_offsets_col->plut), ccb_offsets_col->plut_size);
  }

  msync(ccb_file_contents, ccb_file_size, MS_SYNC);
  munmap(ccb_file_contents, ccb_file_size);
  close(ccb_file_fd);
}

static void process_cary(const struct cary_header *const cary_header) {
  uint32_t i;
  struct ccb_offsets_list *ccb_offsets_list = ccb_offsets_list_create(cary_header);

  current_ary_num += 1;
  current_cel_num = 0;

  ccb_offsets_list_fill(ccb_offsets_list, cary_header);
  ccb_offsets_list_sort(ccb_offsets_list);
  ccb_offsets_list_fill_sizes(ccb_offsets_list, U32SWAP(cary_header->file_size));

  for (i = 0; i < ccb_offsets_list->count; i++) {
    const struct ccb_offsets_col *const ccb_offsets_col = ccb_offsets_list->offsets_col + i;

    process_ccb((const uint8_t *const)cary_header, ccb_offsets_col);
  }

  ccb_offsets_list_destroy(ccb_offsets_list);
}

static void process_stream_file_contents(const uint8_t *const data, size_t size) {
  size_t i;

  for (i = 0; i < size; i++) {
    const struct cary_header *const cary_header = (const struct cary_header *const)(data + i);

    if ((i + sizeof(*cary_header)) > size) {
      break;
    }

    if (cary_header->magic == CARY_MAGIC && cary_header->zero_placeholder == 0) {
      printf("CARY +%zu\n", i);
      process_cary(cary_header);
    }
  }
}

int main(int argc, char const *argv[]) {
  const char *stream_file_path;
  int stream_file_fd;
  void *stream_file_contents;
  struct stat stream_file_stat;

  if (argc < 2) {
    printf("Argument required: path to stream file\n");
    return EXIT_FAILURE;
  }

  stream_file_path = argv[1];

  if (stat(stream_file_path, &stream_file_stat) != 0) {
    perror("Stream file stat error:");
    return EXIT_FAILURE;
  }

  stream_file_fd = open(stream_file_path, O_RDONLY);

  if (stream_file_fd <= STDERR_FILENO) {
    perror("Stream file open error:");
    return EXIT_FAILURE;
  }

  stream_file_contents = mmap(NULL, stream_file_stat.st_size, PROT_READ, MAP_SHARED, stream_file_fd, 0);
  process_stream_file_contents(stream_file_contents, stream_file_stat.st_size);
  munmap(stream_file_contents, stream_file_stat.st_size);
  close(stream_file_fd);

  return EXIT_SUCCESS;
}
